from LnkParse3.target.lnk_target_base import LnkTargetBase


class Printers(LnkTargetBase):
    # TODO Not implemented
    def __init__(self, *args, **kwargs):
        self.name = "Printers"
        return super().__init__(*args, **kwargs)

    def as_item(self):
        item = super().as_item()
        return item



from struct import unpack
import warnings

from LnkParse3.target_factory import TargetFactory

"""
LINKTARGET_IDLIST:
An optional LinkTargetIDList structure (section 2.2), which specifies the
target of the link. The presence of this structure is specified by the
HasLinkTargetIDList bit (LinkFlags section 2.1.1) in the ShellLinkHeader.

------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|         IDListSize           |         IDList (variable)       |
|-------------------------------                                 |
|                             ...                                |
------------------------------------------------------------------
"""


class LnkTargets:
    SIZE_OF_ID_LIST_SIZE = 2

    def __init__(self, indata=None, cp=None):
        self._targets = {}
        self.cp = cp
        self._raw = indata

        start = self.SIZE_OF_ID_LIST_SIZE
        end = self.size()
        self._raw_targets = self._raw[start:end]

    def __iter__(self):
        return self._id_lists()

    def size(self):
        """
        Including IDListSize section itself
        """
        return self.id_list_size() + self.SIZE_OF_ID_LIST_SIZE

    def id_list_size(self):
        """IDListSize (2 bytes):
        The size, in bytes, of the IDList field.
        """
        start, end = 0, self.SIZE_OF_ID_LIST_SIZE
        size = unpack("<H", self._raw[start:end])[0]
        return size

    def _id_lists(self):
        """ItemIDList (variable):
        An array of zero or more ItemID structures (section 2.2.2), which
        contains the item ID list. An IDList structure conforms to the
        following ABNF [RFC5234]:

            IDLIST = *ITEMID TERMINALID

        ------------------------------------------------------------------
        |     0-7b     |     8-15b     |     16-23b     |     24-31b     |
        ------------------------------------------------------------------
        |                       ItemIDList (variable)                    |
        ------------------------------------------------------------------
        |                             ...                                |
        |-----------------------------------------------------------------
        |         TerminalID           |
        --------------------------------
        """
        rest = self._raw_targets
        while rest:
            factory = TargetFactory(indata=rest)
            target_class = factory.target_class()

            if not target_class:
                # Empty or unknown target object.
                break

            target = target_class(indata=rest, cp=self.cp)

            size = factory.item_size()
            rest = rest[size:]
            yield target

    def as_list(self):
        res = []
        for target in self:
            try:
                res.append(target.as_item())
            except KeyError as e:
                msg = f"Error while parsing TargetID `{target.name}` (KeyError {e})"
                warnings.warn(msg)
                continue
        return res



from struct import unpack
import warnings

from LnkParse3.target.unknown import Unknown
from LnkParse3.target.root_folder import RootFolder
from LnkParse3.target.my_computer import MyComputer
from LnkParse3.target.shell_fs_folder import ShellFSFolder
from LnkParse3.target.network_location import NetworkLocation
from LnkParse3.target.compressed_folder import CompressedFolder
from LnkParse3.target.internet import Internet
from LnkParse3.target.control_panel import ControlPanel
from LnkParse3.target.printers import Printers
from LnkParse3.target.common_places_folder import CommonPlacesFolder
from LnkParse3.target.users_files_folder import UsersFilesFolder


class TargetFactory:
    # https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#3-type-indicator-based-shell-items
    SHELL_ITEM_CLASSES = {
        0x00: Unknown,
        0x01: Unknown,
        0x17: Unknown,
        0x1E: RootFolder,
        0x1F: RootFolder,
        0x20: MyComputer,
        0x30: ShellFSFolder,
        0x40: NetworkLocation,
        0x52: CompressedFolder,
        0x61: Internet,
        0x70: ControlPanel,
        0x71: ControlPanel,
        0x72: Printers,
        0x73: CommonPlacesFolder,
        0x74: UsersFilesFolder,
        0x76: Unknown,
        0x80: Unknown,
        0xFF: Unknown,
    }

    def __init__(self, indata):
        self._target = {}
        self._raw = indata

    def item_size(self):
        """ItemIDSize (2 bytes):
        A 16-bit, unsigned integer that specifies the size, in bytes, of the
        ItemID structure, including the ItemIDSize field.
        """
        start, end = 0, 2
        size = unpack("<H", self._raw[start:end])[0]
        return size

    # dup: ./targets/shell_fs_folder.py flags()
    # dup: ./targets/my_computer.py flags()
    def item_type(self):
        """
        Peek item type before creating objects
        """
        start, end = 2, 3
        item_type = unpack("<B", self._raw[start:end])[0]
        return item_type

    def target_class(self):
        if self.item_size() == 0:
            # TerminalID
            return None

        item_type = self.item_type()
        classes = self.SHELL_ITEM_CLASSES

        # TODO: ControlPanelShellItems
        # https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#43-control-panel-shell-items
        # if item_type == 0x00:

        # XXX: Move to table
        try:
            if 0x20 < item_type <= 0x2F:
                target = classes[0x20]
            elif 0x30 < item_type <= 0x3F:
                target = classes[0x30]
            elif 0x40 < item_type <= 0x4F:
                target = classes[0x40]
            else:
                target = classes[item_type]
        except KeyError as e:
            msg = f"Unknown TargetID `{item_type}`"
            warnings.warn(msg)
            return None

        return target



from struct import unpack
from LnkParse3.text_processor import TextProcessor

"""
STRING_DATA:
Zero or more optional StringData structures (section 2.4), which are used to
convey user interface and path identification information. The presence of
these structures is specified by bits (LinkFlags section 2.1.1) in the
ShellLinkHeader.
"""


class StringData:
    def __init__(self, lnk_file, indata=None, cp=None):
        self._raw = indata
        self._data = {}

        self._lnk_file = lnk_file
        self.text_processor = TextProcessor(cp=cp)

        start = 0
        if self._lnk_file.has_name():
            text, length = self.read(self._raw[start:])
            self._data["description"] = text
            start += length

        if self._lnk_file.has_relative_path():
            text, length = self.read(self._raw[start:])
            self._data["relative_path"] = text
            start += length

        if self._lnk_file.has_working_dir():
            text, length = self.read(self._raw[start:])
            self._data["working_directory"] = text
            start += length

        if self._lnk_file.has_arguments():
            text, length = self.read(self._raw[start:])
            self._data["command_line_arguments"] = text
            start += length

        if self._lnk_file.has_icon_location():
            text, length = self.read(self._raw[start:])
            self._data["icon_location"] = text
            start += length

        self._size = start

    def size(self):
        return self._size

    def description(self):
        return self._data.get("description")

    def relative_path(self):
        return self._data.get("relative_path")

    def working_directory(self):
        return self._data.get("working_directory")

    def command_line_arguments(self):
        return self._data.get("command_line_arguments")

    def icon_location(self):
        return self._data.get("icon_location")

    def read(self, binary):
        offset = 2
        char_count = unpack("<H", binary[0:offset])[0]
        length = char_count

        if self._lnk_file.is_unicode():
            self._read = self.text_processor.read_unicode_string
            length *= 2  # UTF-16
        else:
            self._read = self.text_processor.read_string

        text = self._read(binary[offset : offset + length])
        return text, offset + length

    def as_dict(self):
        return {k: v for k, v in self._data.items() if v is not None}



from struct import unpack
from LnkParse3.target.lnk_target_base import LnkTargetBase
from LnkParse3.decorators import uuid

"""
----------------------------------------------------------------------
|             0-7b                 |              8-15b              |
----------------------------------------------------------------------
|   ClassTypeIndicator == 0x1F     | SortIndex == (0x00,0x42,...0x80)|
----------------------------------------------------------------------
|                     <GUID> ShellFolderID                           |
|                             16 B                                   |
----------------------------------------------------------------------
|                         ExtensionBlock                             |
|                              ? B                                   |
----------------------------------------------------------------------
"""


class RootFolder(LnkTargetBase):
    # https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#321-sort-index
    SORT_INDEX = {
        0x00: "Internet Explorer",
        0x42: "Libraries",
        0x44: "Users",
        0x48: "My Documents",
        0x50: "My Computer",
        0x58: "My Networs Places/Network",
        0x60: "Recycle Bin",
        0x68: "Internet Explorer",
        0x70: "Unknown",
        0x80: "My Games",
    }

    def __init__(self, *args, **kwargs):
        self.name = "Root Folder"
        super().__init__(*args, **kwargs)

    def as_item(self):
        item = super().as_item()
        item["sort_index"] = self.sort_index()
        item["guid"] = self.guid()
        return item

    def sort_index(self):
        start, end = 1, 2
        index = unpack("<B", self._raw_target[start:end])[0]
        return self.SORT_INDEX[index]

    @uuid
    def guid(self):
        start, end = 2, 18
        guid = self._raw_target[start:end]
        return guid

    def extension_block(self):
        if self.item_id_size() > 20:
            # TODO: Extension block
            return None
        else:
            return None



from struct import unpack
from LnkParse3.extra.lnk_extra_base import LnkExtraBase
from LnkParse3.decorators import uuid

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|              <u_int32> BlockSize >= 0x0000000C                 |
------------------------------------------------------------------
|            <u_int32> BlockSignature == 0xA0000009              |
------------------------------------------------------------------
|                    <u_int32> StorageSize                       |
------------------------------------------------------------------
|                    Version == 0x53505331                       |
------------------------------------------------------------------
|                      <GUID> FormatID                           |
|                            16 B                                |
------------------------------------------------------------------
|   <vector<MS_OLEPS>> SerializedPropertyValue (see MS-OLEPS)    |
|                             ? B                                |
------------------------------------------------------------------
"""


class Metadata(LnkExtraBase):
    def name(self):
        return "METADATA_PROPERTIES_BLOCK"

    def storage_size(self):
        start, end = 8, 12
        return unpack("<I", self._raw[start:end])[0]

    def version(self):
        start, end = 12, 16
        version = unpack("<I", self._raw[start:end])[0]
        return hex(version)

    @uuid
    def format_id(self):
        start, end = 16, 32
        return self._raw[start:end]

    def as_dict(self):
        tmp = super().as_dict()
        tmp["storage_size"] = self.storage_size()
        tmp["version"] = self.version()
        tmp["format_id"] = self.format_id()
        return tmp

    # TODO:
    def serialized_property_value(self):
        raise NotImplementedError
        """
        if _mpb['format_id'].upper() == 'D5CDD5052E9C101B939708002B2CF9AE':
            # Serialized Property Value (String Name)
            index += 32
            result = []

            while True:
                value = {}
                value['value_size'] = unpack('<I', self.indata[index: index + 4])[0]
                if hex(value['value_size']) == hex(0x0):
                    break
                value['name_size'] = unpack('<I', self.indata[index + 4: index + 8])[0]
                value['name'] = self.read_unicode_string(index + 8)
                value['value'] = ''  # TODO MS-OLEPS

                result.append(value)
                index += 4 + 4 + 2 + value['name_size'] + value['value_size']

            _mpb['serialized_property_value_string'] = result
        else:
            # Serialized Property Value (Integer Name)
            index += 32
            result = []

            while True:
                value = {}
                value['value_size'] = unpack('<I', self.indata[index: index + 4])[0]
                if hex(value['value_size']) == hex(0x0):
                    break
                value['id'] = unpack('<I', self.indata[index + 4: index + 8])[0]
                value['value'] = ''  # TODO MS-OLEPS

                result.append(value)
                index += value['value_size']

            _mpb['serialized_property_value_integer'] = result
        """



from struct import unpack
from LnkParse3.lnk_info import LnkInfo

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|                   <u_int32> VolumeIDSize                       |
------------------------------------------------------------------
|                    <u_int32> DriveType                         |
------------------------------------------------------------------
|                 <u_int32> DriveSerialNumber                    |
------------------------------------------------------------------
|                 <u_int32> VolumeLabelOffset                    |
------------------------------------------------------------------
|        <u_int32> VolumeLabelOffsetUnicode (optional)           |
------------------------------------------------------------------
|                       <u_int32> Data                           |
|                            ? B                                 |
------------------------------------------------------------------
"""


class Local(LnkInfo):
    DRIVE_TYPES = [
        "DRIVE_UNKNOWN",
        "DRIVE_NO_ROOT_DIR",
        "DRIVE_REMOVABLE",
        "DRIVE_FIXED",
        "DRIVE_REMOTE",
        "DRIVE_CDROM",
        "DRIVE_RAMDISK",
    ]

    def location(self):
        return "Local"

    def volume_id_size(self):
        start = self.volume_id_offset()
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def r_drive_type(self):
        start = self.volume_id_offset() + 4
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def drive_serial_number(self):
        start = self.volume_id_offset() + 8
        end = start + 4
        number = unpack("<I", self._raw[start:end])[0]
        return hex(number)

    def volume_label_offset(self):
        start = self.volume_id_offset() + 12
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def drive_type(self):
        if self.r_drive_type() < len(self.DRIVE_TYPES):
            return self.DRIVE_TYPES[self.r_drive_type()]
        else:
            return None

    def _has_volume_label_offset_unicode(self):
        return bool(self.volume_label_offset() == 0x00000014)

    def volume_label(self):
        if self._has_volume_label_offset_unicode():
            return None

        start = self.volume_id_offset() + self.volume_label_offset()

        binary = self._raw[start:]
        text = self.text_processor.read_string(binary)
        return text

    def volume_label_unicode_offset(self):
        if not self._has_volume_label_offset_unicode():
            return None

        start = self.volume_id_offset() + 16
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def volume_label_unicode(self):
        if not self.volume_label_unicode_offset():
            return None

        start = self.volume_id_offset() + self.volume_labe_unicode_offset()

        binary = self._raw[offset:]
        text = self.text_processor.read_unicode_string(binary)
        return text



import json
import os
import unittest
import warnings
from contextlib import redirect_stdout
from io import StringIO

import LnkParse3

TARGET_DIR = os.path.join(os.path.dirname(__file__), 'samples')
JSON_DIR = os.path.join(os.path.dirname(__file__), 'json')


class TestSamples(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        warnings.simplefilter('ignore', category=UserWarning)

    def test_json_print_for_all_samples(self):
        for entry in os.scandir(TARGET_DIR):
            with self.subTest(msg=entry.name):
                with open(entry.path, 'rb') as indata:
                    lnk = LnkParse3.lnk_file(indata)

                mock_stdout = StringIO()
                with redirect_stdout(mock_stdout):
                    lnk.print_json(print_all=True)

                our = json.loads(mock_stdout.getvalue())

                json_path = os.path.join(JSON_DIR, f"{entry.name}.json")
                with open(json_path, 'rb') as fp:
                    their = json.load(fp)

                self.assertDictEqual(our, their)

    def test_unwanted_attributes_are_not_printed_if_not_specified(self):
        with open('tests/samples/microsoft_example', 'rb') as indata:
            lnk = LnkParse3.lnk_file(indata)

        mock_stdout = StringIO()
        with redirect_stdout(mock_stdout):
            lnk.print_json(print_all=False)

        our = json.loads(mock_stdout.getvalue())

        json_path = os.path.join(JSON_DIR, f"microsoft_example_not_all_attributes.json")
        with open(json_path, 'rb') as fp:
            their = json.load(fp)

        self.assertDictEqual(our, their)

    def test_readable_with_network_info(self):
        with open('tests/samples/network_info', 'rb') as indata:
            lnk = LnkParse3.lnk_file(indata)

        mock_stdout = StringIO()
        with redirect_stdout(mock_stdout):
            lnk.print_lnk_file(print_all=True)

        our = mock_stdout.getvalue()

        txt_path = os.path.join(JSON_DIR, f"readable_with_network_info.txt")
        with open(txt_path, 'r') as fp:
            their = fp.read()

        self.assertEqual(our, their)

    def test_readable_with_local_info(self):
        with open('tests/samples/sample', 'rb') as indata:
            lnk = LnkParse3.lnk_file(indata)

        mock_stdout = StringIO()
        with redirect_stdout(mock_stdout):
            lnk.print_lnk_file(print_all=True)

        our = mock_stdout.getvalue()

        txt_path = os.path.join(JSON_DIR, f"readable_with_local_info.txt")
        with open(txt_path, 'r') as fp:
            their = fp.read()

        self.assertEqual(our, their)

    def test_print_shortcut_target_json(self):
        with open('tests/samples/microsoft_example', 'rb') as indata:
            lnk = LnkParse3.lnk_file(indata)

        mock_stdout = StringIO()
        with redirect_stdout(mock_stdout):
            lnk.print_shortcut_target(pjson=True)

        our = json.loads(mock_stdout.getvalue())

        json_path = os.path.join(JSON_DIR, f"shortcut_target_only.json")
        with open(json_path, 'rb') as fp:
            their = json.load(fp)

        self.assertDictEqual(our, their)

    def test_print_shortcut_target_readable(self):
        with open('tests/samples/microsoft_example', 'rb') as indata:
            lnk = LnkParse3.lnk_file(indata)

        mock_stdout = StringIO()
        with redirect_stdout(mock_stdout):
            lnk.print_shortcut_target(pjson=False)

        our = mock_stdout.getvalue()

        json_path = os.path.join(JSON_DIR, f"shortcut_target_only.txt")
        with open(json_path, 'r') as fp:
            their = fp.read()

        self.assertEqual(our, their)


if __name__ == '__main__':
    unittest.main()



from struct import unpack
from LnkParse3.extra.environment import Environment
from LnkParse3.extra.console import Console
from LnkParse3.extra.distributed_tracker import DistributedTracker
from LnkParse3.extra.code_page import CodePage
from LnkParse3.extra.special_folder import SpecialFolder
from LnkParse3.extra.darwin import Darwin
from LnkParse3.extra.icon import Icon
from LnkParse3.extra.shim_layer import ShimLayer
from LnkParse3.extra.metadata import Metadata
from LnkParse3.extra.known_folder import KnownFolder
from LnkParse3.extra.shell_item import ShellItem

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|              <u_int32> BlockSize == 0x00000314                 |
------------------------------------------------------------------
|            <u_int32> BlockSignature == 0xA0000001              |
------------------------------------------------------------------
"""


class ExtraFactory:
    EXTRA_SIGS = {
        "a0000001": Environment,
        "a0000002": Console,
        "a0000003": DistributedTracker,
        "a0000004": CodePage,
        "a0000005": SpecialFolder,
        "a0000006": Darwin,
        "a0000007": Icon,
        "a0000008": ShimLayer,
        "a0000009": Metadata,
        "a000000b": KnownFolder,
        "a000000c": ShellItem,
    }

    def __init__(self, indata):
        self._raw = indata

    def item_size(self):
        start, end = 0, 4
        size = unpack("<I", self._raw[start:end])[0]
        return size

    def _rsig(self):
        start, end = 4, 8
        rsig = unpack("<I", self._raw[start:end])[0]
        return rsig

    def extra_class(self):
        sig = str(hex(self._rsig()))[2:]  # huh?
        return self.EXTRA_SIGS.get(sig)



"""
Structures of the Shell Link Binary File Format can define strings in
fixed-length fields, strings MUST be null-terminated. If a string is smaller
than the size of the field that contains it, the bytes in the field following
the terminating null character are undefined and can have any value. The
undefined bytes MUST NOT be used.
"""
import warnings


class TextProcessor:
    def __init__(self, cp=None):
        self.cp = cp if cp else "cp1252"

    def read_strings(self, binary):
        chars = []

        def _chars_to_string(lst):
            bin_string = b"".join(lst)
            try:
                string = bin_string.decode(self.cp)
            except UnicodeDecodeError as e:
                string = bin_string.decode(self.cp, errors="replace")
                msg = f"Error while decoding string `{string}` ({e})"
                warnings.warn(msg)
            yield string

        for char in binary:
            if char == 0x00:
                yield from _chars_to_string(chars)
                chars = []
            else:
                chars.append(bytes([char]))

        yield from _chars_to_string(chars)

    def read_string(self, binary):
        it = self.read_strings(binary)
        return next(it)

    def read_unicode_strings(self, binary):
        chars = []

        def _chars_to_string(lst):
            bin_string = b"".join(lst)
            try:
                string = bin_string.decode("utf-16le")
            except UnicodeDecodeError as e:
                string = bin_string.decode("utf-16le", errors="replace")
                msg = f"Error while decoding string `{string}` ({e})"
                warnings.warn(msg)
            yield string

        for char in self._2bytes_each(binary):
            if char == b"\x00\x00":
                yield from _chars_to_string(chars)
                chars = []
            else:
                chars.append(char)

        yield from _chars_to_string(chars)

    def read_unicode_string(self, binary):
        it = self.read_unicode_strings(binary)
        return next(it)

    @staticmethod
    def _2bytes_each(binary):
        it = iter(binary)
        while True:
            tmp = None
            try:
                tmp = bytes([next(it)])
                tmp += bytes([next(it)])
                yield tmp
            except StopIteration:
                if tmp:
                    yield tmp
                return None



import warnings
from struct import error as StructError

from LnkParse3.extra_factory import ExtraFactory

"""
EXTRA_DATA:
Zero or more ExtraData structures (section 2.5).
"""


class ExtraData:
    def __init__(self, indata=None, cp=None):
        self.cp = cp
        self._raw = indata

    def __iter__(self):
        return self._iter()

    def _iter(self):
        rest = self._raw
        while rest:
            factory = ExtraFactory(indata=rest)
            size = factory.item_size()

            if not size:
                break

            data, rest = rest[:size], rest[size:]

            cls = factory.extra_class()
            if cls:
                yield cls(indata=data, cp=self.cp)

    def as_dict(self):
        res = {}
        for extra in self:
            try:
                res[extra.name()] = extra.as_dict()
            except (StructError, ValueError) as e:
                msg = "Error while parsing `%s` (%s)" % (extra.name(), e)
                warnings.warn(msg)
                continue
        return res



from datetime import datetime
from datetime import timezone
from struct import unpack
import functools
import sys
import warnings


def must_be(expected):
    def outer(func):
        @functools.wraps(func)
        def inner(self, *args, **kwargs):
            result = func(self, *args, **kwargs)

            if result != expected:
                msg = "%s must be %s: %s" % (func.__name__, expected, result)
                warnings.warn(msg)

            return result

        return inner

    return outer


def uuid(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        binary = func(self, *args, **kwargs)

        # UUID variants
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/49e490b8-f972-45d6-a3a4-99f924998d97
        # Also see Java implementation (mslinks)
        # https://github.com/DmitriiShamrikov/mslinks/blob/master/src/mslinks/data/GUID.java#L51
        d1, d2, d3 = unpack("<LHH", binary[0:8])
        d4, d51, d52 = unpack(">HHI", binary[8:16])

        uuid = "%08X-%04X-%04X-%04X-%04X%08X" % (d1, d2, d3, d4, d51, d52)

        return uuid

    return inner


def _quad_to_hex(quad):
    # An implemetation is based on
    # https://metadataconsulting.blogspot.com/2019/12/CSharp-Convert-a-GUID-to-a-Darwin-Descriptor-and-back.html
    base_85 = "!$%&'()*+,-.0123456789=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{}~"
    i = 5
    ddec = 0
    while i >= 1:
        char = quad[i - 1]
        b85 = base_85.find(char)
        ddec = ddec + b85
        if i > 1:
            ddec = ddec * 85
        i -= 1

    return f"{ddec:08X}"


def packed_uuid(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        text = func(self, *args, **kwargs)

        if text is None:
            return None

        # An implemetation is based on
        # https://metadataconsulting.blogspot.com/2019/12/CSharp-Convert-a-GUID-to-a-Darwin-Descriptor-and-back.html
        quad1 = _quad_to_hex(text[0:5])
        quad2 = _quad_to_hex(text[5:10])
        quad3 = _quad_to_hex(text[10:15])
        quad4 = _quad_to_hex(text[15:20])
        quads = quad1 + quad2 + quad3 + quad4

        d1 = quads[:8]
        d2 = quads[12:16]
        d3 = quads[8:12]
        d41 = quads[22:24]
        d42 = quads[20:22]
        d51 = quads[18:20]
        d52 = quads[16:18]
        d53 = quads[30:32]
        d54 = quads[28:30]
        d55 = quads[26:28]
        d56 = quads[24:26]

        uuid = "%s-%s-%s-%s%s-%s%s%s%s%s%s" % (
            d1,
            d2,
            d3,
            d41,
            d42,
            d51,
            d52,
            d53,
            d54,
            d55,
            d56,
        )

        return uuid

    return inner


def filetime(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        binary = func(self, *args, **kwargs)

        try:
            nanosec = unpack("<q", binary)[0]

            if nanosec == 0:
                # If the date is 0 it means that the file has been created in an
                # application, then saved from it, and never ever opened.
                return None

            epoch_as_filetime = 116444736000000000
            hundreds_of_nanoseconds = 10000000

            timestamp = (nanosec - epoch_as_filetime) / hundreds_of_nanoseconds
            return datetime.fromtimestamp(timestamp, tz=timezone.utc)
        except (OSError, OverflowError, ValueError):
            if sys.version_info < (3, 8, 0):
                # HACK for older versions for bytes.hex()
                # https://docs.python.org/3.9/library/stdtypes.html?highlight=hex#bytes.hex
                iterator = iter(binary.hex())
                invalid_date = " ".join(a + b for a, b in zip(iterator, iterator))
            else:
                invalid_date = binary.hex(" ")
            msg = "Invalid filetime: %s" % invalid_date
            warnings.warn(msg)
            return None

    return inner


def dostime(func):
    r"""
    The DOS date/time format is a bitmask:
    24                16                 8                 0
     +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
     |h|h|h|h|h|m|m|m| |m|m|m|s|s|s|s|s| |Y|Y|Y|Y|Y|Y|Y|M| |M|M|M|D|D|D|D|D|
     +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+ +-+-+-+-+-+-+-+-+
     \________/\____________/\_________/ \____________/\________/\_________/
        hour       minute      second         year       month       day
    The year is stored as an offset from 1980.
    Seconds are stored in two-second increments.
    (So if the "second" value is 15, it actually represents 30 seconds.)
    """
    #
    # Source:
    #   https://github.com/log2timeline/dfdatetime/blob/main/dfdatetime/fat_date_time.py
    #   https://stackoverflow.com/questions/15763259/unix-timestamp-to-fat-timestamp
    #   https://docs.microsoft.com/pl-pl/windows/desktop/api/winbase/nf-winbase-dosdatetimetofiletime
    #   https://github.com/log2timeline/dfdatetime/wiki/Date-and-time-values
    #
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        binary = func(self, *args, **kwargs)

        try:
            dos = unpack("<I", binary)[0]
            # NOTE An alternative solution for DOS conversion is to use
            # `dfdatetime` package. It returns the same date-time for all
            # currently available tests.
            # from dfdatetime.fat_date_time import FATDateTime
            # timestamp = FATDateTime(dos).CopyToPosixTimestamp()
            # return datetime.fromtimestamp(timestamp, tz=timezone.utc)

            if dos == 0:
                # If the date is 0 it means that the file has been created in an
                # application, then saved from it, and never ever opened.
                return None

            ymdhms = (
                ((dos & 0x0000FE00) >> 9) + 1980,
                ((dos & 0x000001E0) >> 5),
                ((dos & 0x0000001F) >> 0),
                ((dos & 0xF8000000) >> 27),
                ((dos & 0x07E00000) >> 21),
                ((dos & 0x001F0000) >> 16) * 2,
            )

            return datetime(*ymdhms, tzinfo=timezone.utc)
        except ValueError:
            if sys.version_info < (3, 8, 0):
                # HACK for older versions for bytes.hex()
                # https://docs.python.org/3.9/library/stdtypes.html?highlight=hex#bytes.hex
                iterator = iter(binary.hex())
                invalid_date = " ".join(a + b for a, b in zip(iterator, iterator))
            else:
                invalid_date = binary.hex(" ")
            msg = "Invalid dostime: %s" % invalid_date
            warnings.warn(msg)
            return None

    return inner



from LnkParse3.target.lnk_target_base import LnkTargetBase


# https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#36-compressed-folder-shell-item
class CompressedFolder(LnkTargetBase):
    # TODO Not implemented
    def __init__(self, *args, **kwargs):
        self.name = "Compressed folder"
        return super().__init__(*args, **kwargs)

    def as_item(self):
        item = super().as_item()
        return item



from struct import unpack
from LnkParse3.decorators import must_be
from LnkParse3.decorators import uuid
from LnkParse3.decorators import filetime
from LnkParse3.exceptions import LnkParserError

"""
SHELL_LINK_HEADER:
A ShellLinkHeader structure (section 2.1), which contains identification
information, timestamps, and flags that specify the presence of optional
structures.

------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|             <u_int32> HeaderSize == 0x0000004C                 |
------------------------------------------------------------------
|   <CSID> LinkCLSID == 00021401-0000-0000-C000-000000000046     |
|                            16 B                                |
------------------------------------------------------------------
|                     <flags> LinkFlags                          |
------------------------------------------------------------------
|                   <flags> FileAttributes                       |
------------------------------------------------------------------
|                  <FILETIME> CreationTime                       |
|                            16 B                                |
------------------------------------------------------------------
|                   <FILETIME> AccessTime                        |
|                            16 B                                |
------------------------------------------------------------------
|                   <FILETIME> WriteTime                         |
|                            16 B                                |
------------------------------------------------------------------
|                   <u_int32> FileSize                           |
------------------------------------------------------------------
|                   <int32> IconIndex                            |
------------------------------------------------------------------
|                 <u_int32> ShowCommand                          |
------------------------------------------------------------------
|   <HotKeyFlags> HotKey      |            Reserved1             |
------------------------------------------------------------------
|                        Reserved2                               |
------------------------------------------------------------------
|                        Reserved3                               |
------------------------------------------------------------------
"""


class LnkHeader:
    # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-showwindow
    WINDOW_STYLES = {
        1: "SW_SHOWNORMAL",
        3: "SW_SHOWMAXIMIZED",
        7: "SW_SHOWMINNOACTIVE",
    }

    HOTKEY_VALUES_HIGH = {
        b"\x00": "UNSET",
        b"\x01": "SHIFT",
        b"\x02": "CONTROL",
        b"\x04": "ALT",
    }

    HOTKEY_VALUES_LOW = {
        b"\x00": "UNSET",
        b"\x30": "0",
        b"\x31": "1",
        b"\x32": "2",
        b"\x33": "3",
        b"\x34": "4",
        b"\x35": "5",
        b"\x36": "6",
        b"\x37": "7",
        b"\x38": "8",
        b"\x39": "9",
        b"\x41": "A",
        b"\x42": "B",
        b"\x43": "C",
        b"\x44": "D",
        b"\x45": "E",
        b"\x46": "F",
        b"\x47": "G",
        b"\x48": "H",
        b"\x49": "I",
        b"\x4A": "J",
        b"\x4B": "K",
        b"\x4C": "L",
        b"\x4D": "M",
        b"\x4E": "N",
        b"\x4F": "O",
        b"\x50": "P",
        b"\x51": "Q",
        b"\x52": "R",
        b"\x53": "S",
        b"\x54": "T",
        b"\x55": "U",
        b"\x56": "V",
        b"\x57": "W",
        b"\x58": "X",
        b"\x59": "Y",
        b"\x5A": "Z",
        b"\x70": "F1",
        b"\x71": "F2",
        b"\x72": "F3",
        b"\x73": "F4",
        b"\x74": "F5",
        b"\x75": "F6",
        b"\x76": "F7",
        b"\x77": "F8",
        b"\x78": "F9",
        b"\x79": "F10",
        b"\x7A": "F11",
        b"\x7B": "F12",
        b"\x7C": "F13",
        b"\x7D": "F14",
        b"\x7E": "F15",
        b"\x7F": "F16",
        b"\x80": "F17",
        b"\x81": "F18",
        b"\x82": "F19",
        b"\x83": "F20",
        b"\x84": "F21",
        b"\x85": "F22",
        b"\x86": "F23",
        b"\x87": "F24",
        b"\x90": "NUM_LOCK",
        b"\x91": "SCROLL_LOCK",
    }

    LINK_FLAG_MASK = {  # {{{
        # LinkTargetIDList structure (section 2.2) MUST follow the
        # ShellLinkHeader. If this bit is not set, this structure MUST NOT
        # be present.
        0x00000001: "HasTargetIDList",
        # The shell link is saved with link information. If this bit is set,
        # a LinkInfo structure (section 2.3) MUST be present. If this bit
        # is not set, this structure MUST NOT be present.
        0x00000002: "HasLinkInfo",
        # The shell link is saved with a name string. If this bit is set,
        # a NAME_STRING StringData structure (section 2.4) MUST be present.
        # If this bit is not set, this structure MUST NOT be present.
        0x00000004: "HasName",
        # The shell link is saved with a relative path string. If this bit
        # is set, a RELATIVE_PATH StringData structure (section 2.4) MUST
        # be present. If this bit is not set, this structure MUST NOT be
        # present.
        0x00000008: "HasRelativePath",
        # The shell link is saved with a working directory string. If this
        # bit is set, a WORKING_DIR StringData structure (section 2.4) MUST
        # be present. If this bit is not set, this structure MUST NOT be
        # present.
        0x00000010: "HasWorkingDir",
        # The shell link is saved with command line arguments. If this bit
        # is set, a COMMAND_LINE_ARGUMENTS StringData structure (section
        # 2.4) MUST be present. If this bit is not set, this structure MUST
        # NOT be present.
        0x00000020: "HasArguments",
        # The shell link is saved with an icon location string. If this bit
        # is set, an ICON_LOCATION StringData structure (section 2.4) MUST
        # be present. If this bit is not set, this structure MUST NOT be
        # present.
        0x00000040: "HasIconLocation",
        # The shell link contains Unicode encoded strings. This bit SHOULD
        # be set. If this bit is set, the StringData section contains
        # Unicode-encoded strings; otherwise, it contains strings that are
        # encoded using the system default code page.
        0x00000080: "IsUnicode",
        # The LinkInfo structure (section 2.3) is ignored.
        0x00000100: "ForceNoLinkInfo",
        # The shell link is saved with an EnvironmentVariableDataBlock
        # (section 2.5.4).
        0x00000200: "HasExpString",
        # The target is run in a separate virtual machine when launching
        # a link target that is a 16-bit application.
        0x00000400: "RunInSeparateProcess",
        # TODO: Unused1
        # A bit that is undefined and MUST be ignored.
        0x00000800: "Reserved0",
        # The shell link is saved with a DarwinDataBlock (section 2.5.3).
        0x00001000: "HasDarwinID",
        # The application is run as a different user when the target of the
        # shell link is activated.
        0x00002000: "RunAsUser",
        # The shell link is saved with an IconEnvironmentDataBlock (section
        # 2.5.5).
        0x00004000: "HasExpIcon",
        # The file system location is represented in the shell namespace
        # when the path to an item is parsed into an IDList.
        0x00008000: "NoPidlAlias",
        # TODO: Unused2
        # A bit that is undefined and MUST be ignored.
        0x00010000: "Reserved1",
        # The shell link is saved with a ShimDataBlock (section 2.5.8).
        0x00020000: "RunWithShimLayer",
        # The TrackerDataBlock (section 2.5.10) is ignored.
        0x00040000: "ForceNoLinkTrack",
        # The shell link attempts to collect target properties and store
        # them in the PropertyStoreDataBlock (section 2.5.7) when the link
        # target is set.
        0x00080000: "EnableTargetMetadata",
        # The EnvironmentVariableDataBlock is ignored.
        0x00100000: "DisableLinkPathTracking",
        # The SpecialFolderDataBlock (section 2.5.9) and the
        # KnownFolderDataBlock (section 2.5.6) are ignored when loading the
        # shell link. If this bit is set, these extra data blocks SHOULD
        # NOT be saved when saving the shell link.
        0x00200000: "DisableKnownFolderTracking",
        # If the link has a KnownFolderDataBlock (section 2.5.6), the
        # unaliased form of the known folder IDList SHOULD be used when
        # translating the target IDList at the time that the link is
        # loaded.
        0x00400000: "DisableKnownFolderAlias",
        # Creating a link that references another link is enabled.
        # Otherwise, specifying a link as the target IDList SHOULD NOT be
        # allowed.
        0x00800000: "AllowLinkToLink",
        # When saving a link for which the target IDList is under a known
        # folder, either the unaliased form of that known folder or the
        # target IDList SHOULD be used.
        0x01000000: "UnaliasOnSave",
        # The target IDList SHOULD NOT be stored; instead, the path
        # specified in the 2.1.2 FileAttributesFlags
        # EnvironmentVariableDataBlock (section 2.5.4) SHOULD be used to
        # refer to the target.
        0x02000000: "PreferEnvironmentPath",
        # When the target is a UNC name that refers to a location on
        # a local machine, the local path IDList in the
        # PropertyStoreDataBlock (section 2.5.7) SHOULD be stored, so it
        # can be used when the link is loaded on the local machine.
        0x04000000: "KeepLocalIDListForUNCTarget",
    }  # }}}

    FILE_FLAG_MASK = {  # {{{
        # The file or directory is read-only. For a file, if this bit is set,
        # applications can read the file but cannot write to it or delete it.
        # For a directory, if this bit is set, applications cannot delete the
        # directory.
        0x00000001: "FILE_ATTRIBUTE_READONLY",
        # The file or directory is hidden. If this bit is set, the file or
        # folder is not included in an ordinary directory listing.
        0x00000002: "FILE_ATTRIBUTE_HIDDEN",
        # The file or directory is part of the operating system or is used
        # exclusively by the operating system.
        0x00000004: "FILE_ATTRIBUTE_SYSTEM",
        # TODO: Reserved1
        # A bit that MUST be zero.
        0x00000008: "Reserved, not used by the LNK format",
        # The link target is a directory instead of a file.
        0x00000010: "FILE_ATTRIBUTE_DIRECTORY",
        # The file or directory is an archive file. Applications use this flag
        # to mark files for backup or removal.
        0x00000020: "FILE_ATTRIBUTE_ARCHIVE",
        # A bit that MUST be zero.
        0x00000040: "FILE_ATTRIBUTE_DEVICE",
        # The file or directory has no other flags set. If this bit is 1, all
        # other bits in this structure MUST be clear.
        0x00000080: "FILE_ATTRIBUTE_NORMAL",
        # The file is being used for temporary storage.
        0x00000100: "FILE_ATTRIBUTE_TEMPORARY",
        # The file is a sparse file.
        0x00000200: "FILE_ATTRIBUTE_SPARSE_FILE",
        # The file or directory has an associated reparse point.
        0x00000400: "FILE_ATTRIBUTE_REPARSE_POINT",
        # The file or directory is compressed. For a file, this means that all
        # data in the file is compressed. For a directory, this means that
        # compression is the default for newly created files and
        # subdirectories.
        0x00000800: "FILE_ATTRIBUTE_COMPRESSED",
        # The data of the file is not immediately available.
        0x00001000: "FILE_ATTRIBUTE_OFFLINE",
        # The contents of the file need to be indexed.
        0x00002000: "FILE_ATTRIBUTE_NOT_CONTENT_INDEXED",
        # The file or directory is encrypted. For a file, this means that all
        # data in the file is encrypted. For a directory, this means that
        # encryption is the default for newly created files and subdirectories.
        0x00004000: "FILE_ATTRIBUTE_ENCRYPTED",
        # The directory or user data stream is configured with integrity
        # (only supported on ReFS volumes).
        0x00008000: "FILE_ATTRIBUTE_INTEGRITY_STREAM",
        # Is virtual
        0x00010000: "FILE_ATTRIBUTE_VIRTUAL",
        # The user data stream not to be read by the background data
        # integrity scanner (AKA scrubber).
        0x00020000: "FILE_ATTRIBUTE_NO_SCRUB_DATA",
    }  # }}}

    def __init__(self, fhandle=None, indata=None):
        if fhandle:
            self._raw = fhandle.read()
        elif indata:
            self._raw = indata
        else:
            raise LnkParserError(
                "Both `LnkHeader` arguments `fhandle` and `indata` are evalued as `None`"
            )

        self._lnk_header = {}
        self._stash = {}

        self._raw = self._raw[: self.size()]

    @must_be(int("0x0000004C", 16))
    def size(self):
        """HeaderSize (4 bytes):
        The size, in bytes, of this structure.
        This value MUST be 0x0000004C.
        """
        start, end = 0, 4
        size = unpack("<I", self._raw[start:end])[0]
        return size

    @must_be("00021401-0000-0000-C000-000000000046")
    @uuid
    def link_cls_id(self):
        """LinkCLSID (16 bytes):
        A class identifier (CLSID).
        This value MUST be 00021401-0000-0000-C000-000000000046.
        """
        return self._raw[4:20]

    def guid(self):
        return self.link_cls_id()

    def r_link_flags(self):
        """LinkFlags (4 bytes):
        A LinkFlags structure (section 2.1.1) that specifies information about
        the shell link and the presence of optional portions of the structure.
        """
        start, end = 20, 24
        flags = unpack("<I", self._raw[start:end])[0]
        return flags

    def link_flags(self):
        """
        The LinkFlags structure defines bits that specify which shell link
        structures are present in the file format after the ShellLinkHeader
        structure (section 2.1).
        """
        items = sorted(self.LINK_FLAG_MASK.items())
        flag = self.r_link_flags()
        return [key for mask, key in items if bool(flag & mask)]

    def r_file_flags(self):
        """FileAttributes (4 bytes):
        A FileAttributesFlags structure (section 2.1.2) that specifies
        information about the link target.
        """
        start, end = 24, 28
        flags = unpack("<I", self._raw[start:end])[0]
        return flags

    def file_flags(self):
        """
        The FileAttributesFlags structure defines bits that specify the file
        attributes of the link target, if the target is a file system item.
        File attributes can be used if the link target is not available, or if
        accessing the target would be inefficient. It is possible for the
        target items attributes to be out of sync with this value.
        """
        items = sorted(self.FILE_FLAG_MASK.items())
        flag = self.r_file_flags()
        return [key for mask, key in items if bool(flag & mask)]

    @filetime
    def creation_time(self):
        """CreationTime (8 bytes):
        A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the
        creation time of the link target in UTC (Coordinated Universal Time).
        If the value is zero, there is no creation time set on the link target.
        """
        start, end = 28, 36
        return self._raw[start:end]

    @filetime
    def access_time(self):
        """AccessTime (8 bytes):
        A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the
        access time of the link target in UTC (Coordinated Universal Time). If
        the value is zero, there is no access time set on the link target.
        """
        start, end = 36, 44
        return self._raw[start:end]

    @filetime
    def write_time(self):
        """WriteTime (8 bytes):
        A FILETIME structure ([MS-DTYP] section 2.3.3) that specifies the write
        time of the link target in UTC (Coordinated Universal Time). If the
        value is zero, there is no write time set on the link target.
        """
        start, end = 44, 52
        return self._raw[start:end]

    def file_size(self):
        """FileSize (4 bytes):
        A 32-bit unsigned integer that specifies the size, in bytes, of the
        link target. If the link target file is larger than 0xFFFFFFFF, this
        value specifies the least significant 32 bits of the link target file
        size.
        """
        start, end = 52, 56
        size = unpack("<I", self._raw[start:end])[0]
        return size

    def icon_index(self):
        """IconIndex (4 bytes):
        A 32-bit signed integer that specifies the index of an icon within
        a given icon location.
        """
        start, end = 56, 60
        index = unpack("<i", self._raw[start:end])[0]
        return index

    # TODO: rename to show_command
    def window_style(self):
        """ShowCommand (4 bytes):
        A 32-bit unsigned integer that specifies the expected window state of
        an application launched by the link. This value SHOULD be one of the
        following.

        * SW_SHOWNORMAL 0x00000001
            The application is open and its window is open in a normal fashion.

        * SW_SHOWMAXIMIZED 0x00000003
            The application is open, and keyboard focus is given to the
            application, but its window is not shown.

        * SW_SHOWMINNOACTIVE 0x00000007
            The application is open, but its window is not shown. It is not
            given the keyboard focus.

        All other values MUST be treated as SW_SHOWNORMAL.
        """
        start, end = 60, 64
        style = unpack("<i", self._raw[start:end])[0]
        fallback = self.WINDOW_STYLES[1]
        return self.WINDOW_STYLES.get(style, fallback)

    # TODO: See _raw_hot_key
    def hot_key(self):
        start, end = 64, 66
        hot_key = self._raw[start:end]
        b_low, b_high = unpack("<cc", hot_key)

        high = self.HOTKEY_VALUES_HIGH.get(b_high)
        low = self.HOTKEY_VALUES_LOW.get(b_low)

        if high and low:
            return "%s - %s {0x%s}" % (high, low, hot_key.hex())
        else:
            return hot_key.hex()

    def raw_hot_key(self):
        """HotKey (2 bytes):
        A HotKeyFlags structure (section 2.1.3) that specifies the keystrokes
        used to launch the application referenced by the shortcut key. This
        value is assigned to the application after it is launched, so that
        pressing the key activates that application.
        """
        start, end = 64, 66
        raw = unpack("<H", self._raw[start:end])[0]
        return raw

    # TODO: rename to reserved1
    @must_be(0)
    def reserved0(self):
        """Reserved1 (2 bytes):
        A value that MUST be zero.
        """
        start, end = 66, 68
        raw = unpack("<H", self._raw[start:end])[0]
        return raw

    # TODO: rename to reserved2
    @must_be(0)
    def reserved1(self):
        """Reserved2 (4 bytes):
        A value that MUST be zero.
        """
        start, end = 68, 72
        raw = unpack("<I", self._raw[start:end])[0]

        return raw

    # TODO: rename to reserved3
    @must_be(0)
    def reserved2(self):
        """Reserved3 (4 bytes):
        A value that MUST be zero.
        """
        start, end = 72, 76
        raw = unpack("<I", self._raw[start:end])[0]

        return raw



from LnkParse3.decorators import uuid, packed_uuid
from LnkParse3.extra.lnk_extra_base import LnkExtraBase

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|              <u_int32> BlockSize == 0x00000314                 |
------------------------------------------------------------------
|            <u_int32> BlockSignature == 0xA0000006              |
------------------------------------------------------------------
|                    <str> DarwinDataAnsi                        |
|                           260 B                                |
------------------------------------------------------------------
|               <unicode_str> DarwinDataUnicode                  |
|                           520 B                                |
------------------------------------------------------------------

DarwinData consists of {Product-Code, Feature Key, Component Code}. It is stored
in a compressed format, e.g. "w_1^VX!!!!!!!!!MKKSkEXCELFiles>tW{~$4Q]c@II=l2xaTO5Z",
which can results into
{91120000-0030-0000-0000-0000000ff1ce}EXCELFiles{0638c49d-bb8b-4cd1-b191-052e8f325736}.

There are four variants according to
https://community.broadcom.com/symantecenterprise/viewdocument/working-with-darwin-descriptors:
1. {compressed product code}{feature name}>{compressed component ID}
2. {compressed product code}>{compressed component ID}
3. {compressed product code}{feature name}<
4. {compressed product code}<

See http://www.laurierhodes.info/?q=node/34 or
https://metadataconsulting.blogspot.com/2019/12/CSharp-Convert-a-GUID-to-a-Darwin-Descriptor-and-back.html
or https://web.archive.org/web/20080323160816/http://support.microsoft.com/kb/243630.
"""


class Darwin(LnkExtraBase):
    def name(self):
        return "DARWIN_BLOCK"

    def darwin_data_ansi(self):
        start = 8
        end = start + 260
        binary = self._raw[start:end]
        text = self.text_processor.read_string(binary)
        return text

    def darwin_data_unicode(self):
        start = 268
        end = start + 520
        binary = self._raw[start:end]
        text = self.text_processor.read_unicode_string(binary)
        return text

    @packed_uuid
    def product_code_id(self):
        data = self.darwin_data_unicode()
        start, end = 0, 20
        text = data[start:end]
        return text

    def feature_name(self):
        data = self.darwin_data_unicode()
        start = 20
        # Search for a termiantor sign which can be either `<` or `>`.
        # None of these characters should be located in a packed GUID.
        # If the character is not found, `find` returns `-1`, i.e. by using max
        # we want to get the one which is present.
        # An absence of both characters leads to an exception.
        terminator = max(data.find(">"), data.find("<"))
        if terminator == start:
            # If the terminator is found but is the same as the start position,
            # there is no feature name.
            return None
        end = terminator
        text = data[start:end]
        return text

    @packed_uuid
    def component_id(self):
        data = self.darwin_data_unicode()
        terminator = data.find(">")
        if terminator == -1:
            # Set the ID to `None` if `terminator` is not found.
            return None
        start = terminator + 1
        text = data[start:]
        return text

    def as_dict(self):
        tmp = super().as_dict()
        tmp["darwin_data_ansi"] = self.darwin_data_ansi()
        tmp["darwin_data_unicode"] = self.darwin_data_unicode()
        tmp["product_code_id"] = self.product_code_id()
        tmp["feature_name"] = self.feature_name()
        tmp["component_id"] = self.component_id()
        return tmp



from struct import unpack
from LnkParse3.target.lnk_target_base import LnkTargetBase

"""
----------------------------------------------------------------------
|              0-7b              |               8-15b               |
----------------------------------------------------------------------
| ClassTypeIndicator == 0x40-0x4F|            UnknownValue           |
----------------------------------------------------------------------
|                          ContentFlags                              |
----------------------------------------------------------------------
|                         <str> Location                             |
|                              ? B                                   |
----------------------------------------------------------------------
|                        <str> Description                           |
|                              ? B                                   |
----------------------------------------------------------------------
|                         <str> Comments                             |
|                              ? B                                   |
----------------------------------------------------------------------
|                         <str> Unknown                              |
|                              ? B                                   |
----------------------------------------------------------------------
"""


# https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#35-network-location-shell-item
class NetworkLocation(LnkTargetBase):
    def __init__(self, *args, **kwargs):
        self.name = "Network location"
        super().__init__(*args, **kwargs)

        it = self._string_data()
        self._location = next(it)

        self._description = None
        if self._has_description():
            self._description = next(it)

        self._comments = None
        if self._has_comments():
            self._comments = next(it)

    def as_item(self):
        item = super().as_item()
        item["flags"] = self.flags()
        item["content_flags"] = self.content_flags()
        item["location"] = self.location()
        return item

    # TODO: rename to class_type_indicator
    def flags(self):
        start, end = 0, 1
        flags = unpack("<B", self._raw_target[start:end])[0]
        return self.SHELL_ITEM_SHEL_FS_FOLDER[flags & 0x0F]

    def content_flags(self):
        """
        0x40 ⇒ has comments
        0x80 ⇒ has description
        """
        start, end = 2, 6
        flags = unpack("<I", self._raw_target[start:end])[0]
        return flags

    def _has_comments(self):
        return bool(self.content_flags() & 0x40)

    def _has_description(self):
        return bool(self.content_flags() & 0x80)

    def _string_data(self):
        start = 6
        binary = self._raw_target[start:]
        return self.text_processor.read_strings(binary)

    def location(self):
        """
        Contains the network name or UNC path
        ASCII string with end-of-string character
        """
        return self._location

    def description(self):
        """
        ASCII string with end-of-string character
        """
        return self._description

    def comments(self):
        """
        ASCII string with end-of-string character
        """
        return self._comments



from LnkParse3.info.local import Local
from LnkParse3.info.network import Network


class InfoFactory:
    def __init__(self, lnk_info):
        self._lnk_info = lnk_info

    def _volume_id_and_local_base_path(self):
        """
        If set, the VolumeID and LocalBasePath fields are present, and their
        locations are specified by the values of the VolumeIDOffset and
        LocalBasePathOffset fields, respectively. If the value of the
        LinkInfoHeaderSize field is greater than or equal to 0x00000024, the
        LocalBasePathUnicode field is present, and its location is specified by
        the value of the LocalBasePathOffsetUnicode field.
        If not set, the VolumeID, LocalBasePath, and LocalBasePathUnicode
        fields are not present, and the values of the VolumeIDOffset and
        LocalBasePathOffset fields are zero. If the value of the
        LinkInfoHeaderSize field is greater than or equal to 0x00000024, the
        value of the LocalBasePathOffsetUnicode field is zero.
        """
        return bool(self._lnk_info.flags() & 0x0001)

    def _common_network_relative_link_and_path_suffix(self):
        """
        If set, the CommonNetworkRelativeLink field is present, and its
        location is specified by the value of the
        CommonNetworkRelativeLinkOffset field.
        If not set, the CommonNetworkRelativeLink field is not present, and the
        value of the CommonNetworkRelativeLinkOffset field is zero.
        """
        return bool(self._lnk_info.flags() & 0x0002)

    def info_class(self):
        if self._volume_id_and_local_base_path():
            return Local
        elif self._common_network_relative_link_and_path_suffix():
            return Network
        else:
            return None



from struct import unpack
from LnkParse3.lnk_info import LnkInfo

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|           <u_int32> CommonNetworkRelativeLinkSize              |
------------------------------------------------------------------
|           <flags> CommonNetworkRelativeLinkFlags               |
------------------------------------------------------------------
|                   <u_int32> NetNameOffset                      |
------------------------------------------------------------------
|                  <u_int32> DeviceNameOffset                    |
------------------------------------------------------------------
|                <u_int32> NetworkProviderType                   |
------------------------------------------------------------------
|          <u_int32> NetNameOffsetUnicode (optional)             |
------------------------------------------------------------------
|         <u_int32> DeviceNameOffsetUnicode (optional)           |
------------------------------------------------------------------
|                       <str> NetName                            |
|                            ? B                                 |
------------------------------------------------------------------
|                     <str> DeviceName                           |
|                            ? B                                 |
------------------------------------------------------------------
|          <unicode_str> NetNameUnicode (optional)               |
|                            ? B                                 |
------------------------------------------------------------------
|         <unicode_str> DeviceNameUnicode (optional)             |
|                            ? B                                 |
------------------------------------------------------------------
"""


class Network(LnkInfo):
    NETWORK_PROVIDER_TYPES = {
        "0x1A000": "WNNC_NET_AVID",
        "0x1B000": "WNNC_NET_DOCUSPACE",
        "0x1C000": "WNNC_NET_MANGOSOFT",
        "0x1D000": "WNNC_NET_SERNET",
        "0X1E000": "WNNC_NET_RIVERFRONT1",
        "0x1F000": "WNNC_NET_RIVERFRONT2",
        "0x20000": "WNNC_NET_DECORB",
        "0x21000": "WNNC_NET_PROTSTOR",
        "0x22000": "WNNC_NET_FJ_REDIR",
        "0x23000": "WNNC_NET_DISTINCT",
        "0x24000": "WNNC_NET_TWINS",
        "0x25000": "WNNC_NET_RDR2SAMPLE",
        "0x26000": "WNNC_NET_CSC",
        "0x27000": "WNNC_NET_3IN1",
        "0x29000": "WNNC_NET_EXTENDNET",
        "0x2A000": "WNNC_NET_STAC",
        "0x2B000": "WNNC_NET_FOXBAT",
        "0x2C000": "WNNC_NET_YAHOO",
        "0x2D000": "WNNC_NET_EXIFS",
        "0x2E000": "WNNC_NET_DAV",
        "0x2F000": "WNNC_NET_KNOWARE",
        "0x30000": "WNNC_NET_OBJECT_DIRE",
        "0x31000": "WNNC_NET_MASFAX",
        "0x32000": "WNNC_NET_HOB_NFS",
        "0x33000": "WNNC_NET_SHIVA",
        "0x34000": "WNNC_NET_IBMAL",
        "0x35000": "WNNC_NET_LOCK",
        "0x36000": "WNNC_NET_TERMSRV",
        "0x37000": "WNNC_NET_SRT",
        "0x38000": "WNNC_NET_QUINCY",
        "0x39000": "WNNC_NET_OPENAFS",
        "0X3A000": "WNNC_NET_AVID1",
        "0x3B000": "WNNC_NET_DFS",
        "0x3C000": "WNNC_NET_KWNP",
        "0x3D000": "WNNC_NET_ZENWORKS",
        "0x3E000": "WNNC_NET_DRIVEONWEB",
        "0x3F000": "WNNC_NET_VMWARE",
        "0x40000": "WNNC_NET_RSFX",
        "0x41000": "WNNC_NET_MFILES",
        "0x42000": "WNNC_NET_MS_NFS",
        "0x43000": "WNNC_NET_GOOGLE",
    }

    def location(self):
        return "Network"

    def common_network_relative_link_size(self):
        start = self.common_network_relative_link_offset()
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def common_network_relative_link_flags(self):
        start = self.common_network_relative_link_offset() + 4
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def net_name_offset(self):
        start = self.common_network_relative_link_offset() + 8
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def device_name_offset(self):
        start = self.common_network_relative_link_offset() + 12
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def r_network_provider_type(self):
        start = self.common_network_relative_link_offset() + 16
        end = start + 4
        provider_type = unpack("<I", self._raw[start:end])[0]
        return hex(provider_type)

    def network_provider_type(self):
        # TODO: explain
        if not self.common_network_relative_link_flags() & 0x0002:
            return None
        if self.r_network_provider_type() not in self.NETWORK_PROVIDER_TYPES:
            return None
        return self.NETWORK_PROVIDER_TYPES[self.r_network_provider_type()]

    def net_name_offset_unicode(self):
        if self.net_name_offset() <= 20:
            return None
        start = self.common_network_relative_link_offset() + 20
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def net_name_unicode(self):
        if self.net_name_offset() <= 20:
            return None

        start = self.common_network_relative_link_offset()
        start += self.net_name_offset_unicode()

        binary = self._raw[start:]
        text = self.text_processor.read_unicode_string(binary)
        return text

    def device_name_offset_unicode(self):
        if self.net_name_offset() <= 20:
            return None
        start = self.common_network_relative_link_offset() + 24
        end = start + 4
        return unpack("<I", self._raw[start:end])[0]

    def device_name_unicode(self):
        if self.net_name_offset() <= 20:
            return None

        start = self.common_network_relative_link_offset()
        start += self.device_name_offset_unicode()

        binary = self._raw[start:]
        text = self.text_processor.read_unicode_string(binary)
        return text

    def net_name(self):
        if self.net_name_offset() > 20:
            return None

        start = self.common_network_relative_link_offset()
        start += self.net_name_offset()

        binary = self._raw[start:]
        text = self.text_processor.read_string(binary)
        return text

    def device_name(self):
        if self.net_name_offset() > 20:
            return None

        start = self.common_network_relative_link_offset()
        start += self.device_name_offset()

        binary = self._raw[start:]
        text = self.text_processor.read_string(binary)
        return text



from struct import unpack
from LnkParse3.text_processor import TextProcessor

"""
LINKINFO:
An optional LinkInfo structure (section 2.3), which specifies information
necessary to resolve the link target. The presence of this structure is
specified by the HasLinkInfo bit (LinkFlags section 2.1.1) in the
ShellLinkHeader.

------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|                   <u_int32> LinkInfoSize                       |
------------------------------------------------------------------
|                <u_int32> LinkInfoHeaderSize                    |
------------------------------------------------------------------
|                   <flags> LinkInfoFlags                        |
------------------------------------------------------------------
|                  <u_int32> VolumeIDOffset                      |
------------------------------------------------------------------
|               <u_int32> LocalBasePathOffset                    |
------------------------------------------------------------------
|           <u_int32> CommonNetworkRelativeLinkOffset            |
------------------------------------------------------------------
|              <u_int32> CommonPathSuffixOffset                  |
------------------------------------------------------------------
|        <u_int32> LocalBasePathOffsetUnicode (optional)         |
------------------------------------------------------------------
|       <u_int32> CommonPathSuffixOffsetUnicode (optional)       |
------------------------------------------------------------------
|                 <VolumeID> VolumeID (optional)                 |
|                            ? B                                 |
------------------------------------------------------------------
|                 <str> LocalBasePath (optional)                 |
|                            ? B                                 |
------------------------------------------------------------------
|<CommonNetworkRelativeLink> CommonNetworkRelativeLink (optional)|
|                            ? B                                 |
------------------------------------------------------------------
|               <str> CommonPathSuffix (optional)                |
|                            ? B                                 |
------------------------------------------------------------------
|        <unicode_str> LocalBasePathUnicode (optional)           |
|                            ? B                                 |
------------------------------------------------------------------
|      <unicode_str> CommonPathSuffixUnicode (optional)          |
|                            ? B                                 |
------------------------------------------------------------------
"""


class LnkInfo:
    def __init__(self, indata=None, cp=None):
        self._raw = indata
        self.text_processor = TextProcessor(cp=cp)

    def size(self):
        """LinkInfoSize (4 bytes):
        A 32-bit, unsigned integer that specifies the size, in bytes, of the
        LinkInfo structure. All offsets specified in this structure MUST be
        less than this value, and all strings contained in this structure MUST
        fit within the extent defined by this size.
        """
        start, end = 0, 4
        return unpack("<I", self._raw[start:end])[0]

    def header_size(self):
        """LinkInfoHeaderSize (4 bytes):
        A 32-bit, unsigned integer that specifies the size, in bytes, of the
        LinkInfo header section, which is composed of the LinkInfoSize,
        LinkInfoHeaderSize, LinkInfoFlags, VolumeIDOffset, LocalBasePathOffset,
        CommonNetworkRelativeLinkOffset, CommonPathSuffixOffset fields, and, if
        included, the LocalBasePathOffsetUnicode and
        CommonPathSuffixOffsetUnicode fields

        * 0x0000001C
            Offsets to the optional fields are not specified.

        * 0x00000024 ≤ value
            Offsets to the optional fields are specified.
        """
        start, end = 4, 8
        return unpack("<I", self._raw[start:end])[0]

    def flags(self):
        """LinkInfoFlags (4 bytes):
        Flags that specify whether the VolumeID, LocalBasePath,
        LocalBasePathUnicode, and CommonNetworkRelativeLink fields are present
        in this structure.
        """
        start, end = 8, 12
        return unpack("<I", self._raw[start:end])[0]

    def volume_id_offset(self):
        """VolumeIDOffset (4 bytes):
        A 32-bit, unsigned integer that specifies the location of the VolumeID
        field. If the VolumeIDAndLocalBasePath flag is set, this value is an
        offset, in bytes, from the start of the LinkInfo structure; otherwise,
        this value MUST be zero.
        """
        start, end = 12, 16
        return unpack("<I", self._raw[start:end])[0]

    def local_base_path_offset(self):
        """LocalBasePathOffset (4 bytes):
        A 32-bit, unsigned integer that specifies the location of the
        LocalBasePath field. If the VolumeIDAndLocalBasePath flag is set, this
        value is an offset, in bytes, from the start of the LinkInfo structure;
        otherwise, this value MUST be zero.
        """
        start, end = 16, 20
        return unpack("<I", self._raw[start:end])[0]

    def common_network_relative_link_offset(self):
        """CommonNetworkRelativeLinkOffset (4 bytes):
        A 32-bit, unsigned integer that specifies the location of the
        CommonNetworkRelativeLink field. If the
        CommonNetworkRelativeLinkAndPathSuffix flag is set, this value is an
        offset, in bytes, from the start of the LinkInfo structure; otherwise,
        this value MUST be zero.
        """
        start, end = 20, 24
        return unpack("<I", self._raw[start:end])[0]

    def common_path_suffix_offset(self):
        """CommonPathSuffixOffset (4 bytes):
        A 32-bit, unsigned integer that specifies the location of the
        CommonPathSuffix field. This value is an offset, in bytes, from the
        start of the LinkInfo structure.
        """
        start, end = 24, 28
        return unpack("<I", self._raw[start:end])[0]

    def _has_opt_fields(self):
        """
        Offsets to the optional fields are specified.
        """
        return bool(self.header_size() >= 0x00000024)

    def common_path_suffix(self):
        """CommonPathSuffix (variable):
        A NULL-terminated string, defined by the system default code page,
        which is used to construct the full path to the link item or link
        target by being appended to the string in the LocalBasePath field.
        """
        if not self.common_path_suffix_offset():
            return None

        start = self.common_path_suffix_offset()

        binary = self._raw[start:]
        text = self.text_processor.read_string(binary)
        return text

    def local_base_path_offset_unicode(self):
        """LocalBasePathOffsetUnicode (4 bytes):
        An optional, 32-bit, unsigned integer that specifies the location of
        the LocalBasePathUnicode field. If the VolumeIDAndLocalBasePath flag
        is set, this value is an offset, in bytes, from the start of the
        LinkInfo structure; otherwise, this value MUST be zero. This field
        can be present only if the value of the LinkInfoHeaderSize field is
        greater than or equal to 0x00000024.
        """
        if not self._has_opt_fields():
            return None

        start, end = 28, 32
        return unpack("<I", self._raw[start:end])[0]

    def common_path_suffix_offset_unicode(self):
        """CommonPathSuffixOffsetUnicode (4 bytes):
        An optional, 32-bit, unsigned integer that specifies the location of
        the CommonPathSuffixUnicode field. This value is an offset, in bytes,
        from the start of the LinkInfo structure. This field can be present
        only if the value of the LinkInfoHeaderSize field is greater than or
        equal to 0x00000024.
        """
        if not self._has_opt_fields():
            return None

        start, end = 32, 36
        return unpack("<I", self._raw[start:end])[0]

    def local_base_path(self):
        """LocalBasePath (variable):
        An optional, NULL-terminated string, defined by the system default
        code page, which is used to construct the full path to the link item
        or link target by appending the string in the CommonPathSuffix field.
        This field is present if the VolumeIDAndLocalBasePath flag is set.
        """
        if not self.local_base_path_offset():
            return None

        start = self.local_base_path_offset()

        binary = self._raw[start:]
        text = self.text_processor.read_string(binary)
        return text

    def common_network_relative_link(self):
        """CommonNetworkRelativeLink (variable):
        An optional CommonNetworkRelativeLink structure (section 2.3.2) that
        specifies information about the network location where the link
        target is stored.
        """
        # TODO:
        pass

    def local_base_path_unicode(self):
        """LocalBasePathUnicode (variable):
        An optional, NULL-terminated, Unicode string that is used to construct
        the full path to the link item or link target by appending the string
        in the CommonPathSuffixUnicode field. This field can be present only
        if the VolumeIDAndLocalBasePath flag is set and the value of the
        LinkInfoHeaderSize field is greater than or equal to 0x00000024.
        """
        if not self.local_base_path_offset_unicode():
            return None

        start = self.local_base_path_offset_unicode()

        binary = self._raw[start:]
        text = self.text_processor.read_unicode_string(binary)
        return text

    def common_path_suffix_unicode(self):
        """CommonPathSuffixUnicode (variable):
        An optional, NULL-terminated, Unicode string that is used to construct
        the full path to the link item or link target by appending the string
        in the CommonPathSuffixUnicode field. This field can be present only
        if the VolumeIDAndLocalBasePath flag is set and the value of the
        LinkInfoHeaderSize field is greater than or equal to 0x00000024.
        """
        if not self.common_path_suffix_offset_unicode():
            return None

        start = self.common_path_suffix_offset_unicode() + 4

        binary = self._raw[start:]
        # import pdb;pdb.set_trace()
        text = self.text_processor.read_unicode_string(binary)
        return text



from LnkParse3.target.lnk_target_base import LnkTargetBase
from LnkParse3.decorators import uuid


# https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#38-control-panel-shell-item
class ControlPanel(LnkTargetBase):
    # TODO Not implemented
    def __init__(self, *args, **kwargs):
        self.name = "Control panel"
        return super().__init__(*args, **kwargs)

    @uuid
    def control_panel_item_identifier(self):
        start, end = 14, 30
        return self._raw_target[start:end]

    def as_item(self):
        item = super().as_item()
        item["item_identifier"] = self.control_panel_item_identifier()
        return item



from struct import unpack
from LnkParse3.extra.lnk_extra_base import LnkExtraBase
from LnkParse3.decorators import must_be
from LnkParse3.decorators import uuid

"""
------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|              <u_int32> BlockSize == 0x00000060                 |
------------------------------------------------------------------
|            <u_int32> BlockSignature == 0xA0000003              |
------------------------------------------------------------------
|                      <u_int32> Length                          |
------------------------------------------------------------------
|                      <u_int32> Version                         |
------------------------------------------------------------------
|                       <str> MachineID                          |
|                             16 B                               |
------------------------------------------------------------------
|                    <GUID> DroidVolumeId                        |
|                             16 B                               |
------------------------------------------------------------------
|                     <GUID> DroidFileId                         |
|                             16 B                               |
------------------------------------------------------------------
|                  <GUID> DroidBirthVolumeId                     |
|                             16 B                               |
------------------------------------------------------------------
|                   <GUID> DroidBirthFileId                      |
|                             16 B                               |
------------------------------------------------------------------
"""


class DistributedTracker(LnkExtraBase):
    def name(self):
        return "DISTRIBUTED_LINK_TRACKER_BLOCK"

    @must_be(0x00000058)
    def length(self):
        """Length (4 bytes):
        A 32-bit, unsigned integer that specifies the size of the rest of the
        TrackerDataBlock structure, including this Length field. This value
        MUST be 0x00000058.
        """
        start, end = 8, 12
        length = unpack("<I", self._raw[start:end])[0]
        return length

    @must_be(0x00000000)
    def version(self):
        """Version (4 bytes):
        A 32-bit, unsigned integer. This value MUST be 0x00000000.
        """
        start, end = 12, 16
        version = unpack("<I", self._raw[start:end])[0]
        return version

    def machine_id(self):
        """MachineID (16 bytes):
        A NULL-terminated character string, as defined by
        the system default code
        """
        start = 16
        end = start + 16
        binary = self._raw[start:end]
        text = self.text_processor.read_string(binary)
        return text

    @uuid
    def droid_volume_id(self):
        start, end = 32, 48
        binary = self._raw[start:end]
        return binary

    @uuid
    def droid_file_id(self):
        start, end = 48, 64
        binary = self._raw[start:end]
        return binary

    @uuid
    def droid_birth_volume_id(self):
        start, end = 64, 80
        binary = self._raw[start:end]
        return binary

    @uuid
    def droid_birth_file_id(self):
        start, end = 80, 96
        binary = self._raw[start:end]
        return binary

    def as_dict(self):
        tmp = super().as_dict()
        tmp["length"] = self.length()
        tmp["version"] = self.version()
        tmp["machine_identifier"] = self.machine_id()
        tmp["droid_volume_identifier"] = self.droid_volume_id()
        tmp["droid_file_identifier"] = self.droid_file_id()
        tmp["birth_droid_volume_identifier"] = self.droid_birth_volume_id()
        tmp["birth_droid_file_identifier"] = self.droid_birth_file_id()
        return tmp



from struct import unpack
from LnkParse3.target.lnk_target_base import LnkTargetBase
from LnkParse3.decorators import dostime

"""
----------------------------------------------------------------------
|              0-7b              |               8-15b               |
----------------------------------------------------------------------
| ClassTypeIndicator == 0x30-0x3F|            UnknownValue           |
----------------------------------------------------------------------
|                        <u_int16> FileSize                          |
|                              4 B                                   |
----------------------------------------------------------------------
|           <dos_timestamp> LastModificationDateAndTime              |
|                              4 B                                   |
----------------------------------------------------------------------
|                       FileAttributeFlags                           |
----------------------------------------------------------------------
|                 <str/unicode_str> PrimaryName                      |
|                              ? B                                   |
----------------------------------------------------------------------
|                           UnknownData                              |
|                              ? B                                   |
----------------------------------------------------------------------
"""


# TODO: rename to file_entry
# https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#34-file-entry-shell-item
class ShellFSFolder(LnkTargetBase):
    def __init__(self, *args, **kwargs):
        self.name = "File entry"
        super().__init__(*args, **kwargs)

    def as_item(self):
        item = super().as_item()
        try:
            item["flags"] = self.flags()
            item["file_size"] = self.file_size()
            item["modification_time"] = self.modification_time()
            item["file_attribute_flags"] = self.file_attribute_flags()
            item["primary_name"] = self.primary_name()
        except KeyError:
            # FIXME This try-catch is just a hot-fix.
            # We should probably solve failing attributes in a better way.
            pass
        return item

    # dup: ./my_computer.py flags()
    # dup: ../target_factory.py item_type()
    def flags(self):
        flags = self.class_type_indicator()

        # FIXME: delete masking
        return self.SHELL_ITEM_SHEL_FS_FOLDER[flags & 0x0F]

    def file_size(self):
        start, end = 2, 6
        size = unpack("<I", self._raw_target[start:end])[0]
        return size

    @dostime
    def modification_time(self):
        start, end = 6, 10
        return self._raw_target[start:end]

    def file_attribute_flags(self):
        start, end = 10, 12
        flags = unpack("<H", self._raw_target[start:end])[0]
        return flags

    def primary_name(self):
        start = 12
        binary = self._raw_target[start:]

        if self.has_unicode_strings():
            text = self.text_processor.read_unicode_string(binary)
        else:
            text = self.text_processor.read_string(binary)

        return text

    # https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#341-file-entry-shell-item--pre-windows-xp
    def secondary_name(self):
        # TODO:
        pass

    def shell_folder_identifier(self):
        # TODO:
        pass



from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='LnkParse3',
    version='1.2.0',
    description='Windows Shortcut file (LNK) parser',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Matmaus/LnkParse',
    author='Matmaus',
    author_email='matusjas.work@gmail.com',
    license='MIT',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'lnkparse=LnkParse3.lnk_file:main',
        ],
    },
)



from struct import unpack
from LnkParse3.text_processor import TextProcessor

"""
An ItemID is an element in an IDList structure (section 2.2.1). The data stored
in a given ItemID is defined by the source that corresponds to the location in
the target namespace of the preceding ItemIDs. This data uniquely identifies
the items in that part of the namespace.

------------------------------------------------------------------
|     0-7b     |     8-15b     |     16-23b     |     24-31b     |
------------------------------------------------------------------
|         ItemIDSize           |          Data (variable)        |
------------------------------------------------------------------
|                             ...                                |
------------------------------------------------------------------
"""


class LnkTargetBase:
    SHELL_ITEM_SHEL_FS_FOLDER = {
        0x01: "Is directory",
        0x02: "Is file",
        0x04: "Has Unicode strings",
        0x08: "Unknown",
        0x80: "Has CLSID",
    }

    SIZE_OF_TARGET_SIZE = 2

    def __init__(self, indata=None, cp=None):
        self._target = {}
        self.cp = cp
        self._raw = indata

        self.text_processor = TextProcessor(cp=self.cp)

        start = self.SIZE_OF_TARGET_SIZE
        end = start + self.size()
        self._raw_target = self._raw[start:end]

    def as_item(self):
        return {
            "class": self.name,
        }

    def size(self):
        """ItemIDSize (2 bytes):
        A 16-bit, unsigned integer that specifies the size, in bytes, of the
        ItemID structure, including the ItemIDSize field.
        """
        start, end = 0, 2
        size = unpack("<H", self._raw[start:end])[0]
        return size

    def class_type_indicator(self):
        start, end = 0, 1
        flags = unpack("<B", self._raw_target[start:end])[0]
        return flags

    def has_unicode_strings(self):
        inv = {v: k for k, v in self.SHELL_ITEM_SHEL_FS_FOLDER.items()}
        mask = inv["Has Unicode strings"]
        return bool(self.class_type_indicator() & mask)



from struct import unpack
from LnkParse3.target.lnk_target_base import LnkTargetBase

"""
----------------------------------------------------------------------
|              0-7b              |               8-15b               |
----------------------------------------------------------------------
| ClassTypeIndicator == 0x20-0x2F|            UnknownData            |
----------------------------------                                   |
|                                             ? B                    |
----------------------------------------------------------------------

https://github.com/libyal/libfwsi/blob/master/documentation/Windows%20Shell%20Item%20format.asciidoc#33-volume-shell-item
"""


# TODO: rename to volume_shell_item
class MyComputer(LnkTargetBase):
    def __init__(self, *args, **kwargs):
        self.name = "Volume Item"
        super().__init__(*args, **kwargs)

    def as_item(self):
        item = super().as_item()
        item["flags"] = self.flags()
        item["data"] = self.data()
        return item

    # dup: ./shell_fs_folder.py flags()
    # dup: ../target_factory.py item_type()
    def flags(self):
        flags = self.class_type_indicator()

        # FIXME: delete masking
        # FIXME: hex() is only used here
        return hex(flags & 0x0F)

    def data(self):
        return None



#!/usr/bin/env python3

__description__ = "Windows Shortcut file (LNK) parser"
__author__ = "Matmaus"
__version__ = "1.2.0"

import json
import datetime
import argparse
from subprocess import list2cmdline

from LnkParse3.lnk_header import LnkHeader
from LnkParse3.lnk_targets import LnkTargets
from LnkParse3.lnk_info import LnkInfo
from LnkParse3.info_factory import InfoFactory
from LnkParse3.string_data import StringData
from LnkParse3.extra_data import ExtraData


class LnkFile(object):
    def __init__(self, fhandle=None, indata=None, cp=None):
        if fhandle:
            self.indata = fhandle.read()
        elif indata:
            self.indata = indata

        self.cp = cp

        self.process()

    def has_relative_path(self):
        return bool("HasRelativePath" in self.header.link_flags())

    def has_arguments(self):
        return bool("HasArguments" in self.header.link_flags())

    def is_unicode(self):
        return bool("IsUnicode" in self.header.link_flags())

    def has_name(self):
        return bool("HasName" in self.header.link_flags())

    def has_working_dir(self):
        return bool("HasWorkingDir" in self.header.link_flags())

    def has_icon_location(self):
        return bool("HasIconLocation" in self.header.link_flags())

    def has_target_id_list(self):
        return bool("HasTargetIDList" in self.header.link_flags())

    def has_link_info(self):
        return bool("HasLinkInfo" in self.header.link_flags())

    def force_no_link_info(self):
        return bool("ForceNoLinkInfo" in self.header.link_flags())

    def process(self):
        index = 0

        # Parse header
        self.header = LnkHeader(indata=self.indata)
        index += self.header.size()

        # XXX: json
        self._target_index = index + 2

        # Parse ID List
        self.targets = None
        if self.has_target_id_list():
            self.targets = LnkTargets(indata=self.indata[index:], cp=self.cp)
            index += self.targets.size()

        # Parse Link Info
        self.info = None
        if self.has_link_info() and not self.force_no_link_info():
            info = LnkInfo(indata=self.indata[index:], cp=self.cp)
            info_class = InfoFactory(info).info_class()
            if info_class:
                self.info = info_class(indata=self.indata[index:], cp=self.cp)
                index += self.info.size()

        # Parse String Data
        self.string_data = StringData(self, indata=self.indata[index:], cp=self.cp)
        index += self.string_data.size()

        # Parse Extra Data
        self.extras = ExtraData(indata=self.indata[index:], cp=self.cp)

    def print_lnk_file(self, print_all=False):
        def cprint(text, level=0):
            SPACING = 3
            UNWANTED_TRAITS = ["offset", "reserved", "size"]
            text = str(text)
            if print_all or all(x not in text.lower() for x in UNWANTED_TRAITS):
                print(" " * (level * SPACING) + text)  # add leading spaces

        def nice_id(identifier):
            return identifier.capitalize().replace("_", " ")

        # TODO recursive nice print
        cprint("Windows Shortcut Information:")
        cprint("Header Size: %s" % self.header.size(), 1)
        cprint("Link CLSID: %s" % self.header.link_cls_id(), 1)
        cprint(
            "Link Flags: %s - (%s)"
            % (self.format_linkFlags(), self.header.r_link_flags()),
            1,
        )
        cprint(
            "File Flags: %s - (%s)"
            % (self.format_fileFlags(), self.header.r_file_flags()),
            1,
        )
        cprint("")
        cprint("Creation Timestamp: %s" % (self.header.creation_time()), 1)
        cprint("Modified Timestamp: %s" % (self.header.write_time()), 1)
        cprint("Accessed Timestamp: %s" % (self.header.access_time()), 1)
        cprint("")
        cprint(
            "File Size: %s (r: %s)"
            % (str(self.header.file_size()), str(len(self.indata))),
            1,
        )
        cprint("Icon Index: %s " % (str(self.header.icon_index())), 1)
        cprint("Window Style: %s " % (str(self.header.window_style())), 1)
        cprint("HotKey: %s " % (str(self.header.hot_key())), 1)
        cprint("Reserved0: %s" % self.header.reserved0(), 1)
        cprint("Reserved1: %s" % self.header.reserved1(), 1)
        cprint("Reserved2: %s" % self.header.reserved2(), 1)
        cprint("")

        if self.targets:
            cprint("TARGETS:", 1)
            cprint("Size: %s" % self.targets.id_list_size(), 2)
            cprint("Index: %s" % self._target_index, 2)
            cprint("ITEMS:", 2)
            for target in self.targets.as_list():
                cprint(target["class"], 3)
                for key, value in target.items():
                    if key != "class":
                        cprint(f"{nice_id(key)}: {value}", 4)
            cprint("")

        if self.info:
            cprint("LINK INFO:", 1)
            cprint("Link info size: %s" % self.info.size(), 2)
            cprint("Link info header size: %s" % self.info.header_size(), 2)
            cprint("Link info flags: %s" % self.info.flags(), 2)
            cprint("Volume ID offset: %s" % self.info.volume_id_offset(), 2)
            cprint("Local base path offset: %s" % self.info.local_base_path_offset(), 2)
            cprint(
                "Common network relative link offset: %s"
                % self.info.common_network_relative_link_offset(),
                2,
            )
            cprint(
                "Common path suffix offset: %s" % self.info.common_path_suffix_offset(),
                2,
            )
            if self.info.local_base_path_offset():
                cprint("Local base path: %s" % self.info.local_base_path(), 2)
            if self.info.common_path_suffix_offset():
                cprint("Common path suffix: %s" % self.info.common_path_suffix(), 2)
            if self.info.local_base_path_offset_unicode():
                cprint(
                    "Local base path offset unicode: %s"
                    % self.info.local_base_path_offset_unicode(),
                    2,
                )
                cprint(
                    "Local base unicode: %s" % self.info.local_base_path_unicode(), 2
                )
            if self.info.common_path_suffix_offset_unicode():
                cprint(
                    "Common path suffix offset unicode: %s"
                    % self.info.common_path_suffix_offset_unicode(),
                    2,
                )
                cprint(
                    "Common path suffix unicode: %s"
                    % self.info.common_path_suffix_unicode(),
                    2,
                )
            if type(self.info).__name__ == "Local":
                cprint("LOCAL:", 2)
                cprint("Volume ID size: %s" % self.info.volume_id_size(), 3)
                cprint("Drive type: %s" % self.info.r_drive_type(), 3)
                cprint("Volume label offset: %s" % self.info.volume_label_offset(), 3)
                cprint("Drive serial number: %s" % self.info.drive_serial_number(), 3)
                cprint("Drive type: %s" % self.info.drive_type(), 3)
                cprint("Volume label: %s" % self.info.volume_label(), 3)
                if self.info.common_network_relative_link():
                    cprint(
                        "Common network relative link: %s"
                        % self.info.common_network_relative_link(),
                        3,
                    )
                if self.info.volume_label_unicode_offset():
                    cprint(
                        "Volume label unicode offset: %s"
                        % self.info.volume_label_unicode_offset(),
                        3,
                    )
                    cprint(
                        "Volume label unicode: %s" % self.info.volume_label_unicode(), 3
                    )
            elif type(self.info).__name__ == "Network":
                cprint(
                    "Common network relative link size: %s"
                    % self.info.common_network_relative_link_size(),
                    3,
                )
                cprint(
                    "Common network relative link flags: %s"
                    % self.info.common_network_relative_link_flags(),
                    3,
                )
                cprint("Net name offset: %s" % self.info.net_name_offset(), 3)
                cprint("Device name offset: %s" % self.info.device_name_offset(), 3)
                cprint(
                    "Network provider type: %s" % self.info.r_network_provider_type(), 3
                )
                if self.info.network_provider_type():
                    cprint(
                        "Network provider type: %s" % self.info.network_provider_type(),
                        3,
                    )
                if self.info.net_name_offset_unicode():
                    cprint(
                        "Net name offset unicode: %s"
                        % self.info.net_name_offset_unicode(),
                        3,
                    )
                    cprint("Net name unicode: %s" % self.info.net_name_unicode(), 3)
                if self.info.device_name_offset_unicode():
                    cprint(
                        "Device name offset unicode: %s"
                        % self.info.device_name_offset_unicode(),
                        3,
                    )
                    cprint(
                        "Device name unicode: %s" % self.info.device_name_unicode(), 3
                    )
                if self.info.net_name():
                    cprint("Net name: %s" % self.info.net_name(), 3)
                if self.info.device_name():
                    cprint("Device name: %s" % self.info.device_name(), 3)
            cprint("")

        cprint("DATA", 1)
        for key, value in self.string_data.as_dict().items():
            cprint("%s: %s" % (nice_id(key), value), 2)
        cprint("")

        cprint("EXTRA BLOCKS:", 1)
        for extra_key, extra_value in self.extras.as_dict().items():
            cprint(f"{extra_key}", 2)
            for key, value in extra_value.items():
                cprint(f"{nice_id(key)}: {value}", 3)

    def format_linkFlags(self):
        return " | ".join(self.header.link_flags())

    def format_fileFlags(self):
        return " | ".join(self.header.file_flags())

    # FIXME: Simple concat of path and arguments
    @property
    def lnk_command(self):
        out = []

        if self.has_relative_path():
            relative_path = self.string_data.relative_path()
            out.append(list2cmdline([relative_path]))

        if self.has_arguments():
            out.append(self.string_data.command_line_arguments())

        return " ".join(out)

    def print_shortcut_target(self, pjson=False):
        out = self.lnk_command

        if pjson:
            print(json.dumps({"shortcut_target": out}))
        else:
            print(out)

    def print_json(self, print_all=False):
        res = self.get_json(print_all)

        def _datetime_to_str(obj):
            if isinstance(obj, datetime.datetime):
                return obj.replace(microsecond=0).isoformat()
            return obj

        print(
            json.dumps(
                res,
                indent=4,
                separators=(",", ": "),
                default=_datetime_to_str,
                sort_keys=True,
            )
        )

    def get_json(self, get_all=False):
        res = {
            "header": {
                "guid": self.header.link_cls_id(),
                "r_link_flags": self.header.r_link_flags(),
                "r_file_flags": self.header.r_file_flags(),
                "creation_time": self.header.creation_time(),
                "accessed_time": self.header.access_time(),
                "modified_time": self.header.write_time(),
                "file_size": self.header.file_size(),
                "icon_index": self.header.icon_index(),
                "windowstyle": self.header.window_style(),
                "hotkey": self.header.hot_key(),
                "r_hotkey": self.header.raw_hot_key(),
                "link_flags": self.header.link_flags(),
                "file_flags": self.header.file_flags(),
                "header_size": self.header.size(),
                "reserved0": self.header.reserved0(),
                "reserved1": self.header.reserved1(),
                "reserved2": self.header.reserved2(),
            },
            "data": self.string_data.as_dict(),
            "extra": self.extras.as_dict(),
        }

        if self.targets:
            res["target"] = {
                "size": self.targets.id_list_size(),
                "items": self.targets.as_list(),
                "index": self._target_index,
            }

        res["link_info"] = {}
        if self.info:
            res["link_info"] = {
                "link_info_size": self.info.size(),
                "link_info_header_size": self.info.header_size(),
                "link_info_flags": self.info.flags(),
                "volume_id_offset": self.info.volume_id_offset(),
                "local_base_path_offset": self.info.local_base_path_offset(),
                "common_network_relative_link_offset": self.info.common_network_relative_link_offset(),
                "common_path_suffix_offset": self.info.common_path_suffix_offset(),
            }

            if self.info.local_base_path_offset():
                res["link_info"]["local_base_path"] = self.info.local_base_path()
            if self.info.common_path_suffix_offset():
                res["link_info"]["common_path_suffix"] = self.info.common_path_suffix()
            if self.info.local_base_path_offset_unicode():
                res["link_info"][
                    "local_base_path_offset_unicode"
                ] = self.info.local_base_path_offset_unicode()
                res["link_info"][
                    "local_base_path_unicode"
                ] = self.info.local_base_path_unicode()
            if self.info.common_path_suffix_offset_unicode():
                res["link_info"][
                    "common_path_suffix_offset_unicode"
                ] = self.info.common_path_suffix_offset_unicode()
                res["link_info"][
                    "common_path_suffix_unicode"
                ] = self.info.common_path_suffix_unicode()

            res["link_info"]["location_info"] = {}
            if type(self.info).__name__ == "Local":
                res["link_info"]["location"] = self.info.location()

                res["link_info"]["location_info"] = {
                    "volume_id_size": self.info.volume_id_size(),
                    "r_drive_type": self.info.r_drive_type(),
                    "volume_label_offset": self.info.volume_label_offset(),
                    "drive_serial_number": self.info.drive_serial_number(),
                    "drive_type": self.info.drive_type(),
                    "volume_label": self.info.volume_label(),
                }

                if self.info.common_network_relative_link():
                    res["link_info"]["location_info"][
                        "common_network_relative_link"
                    ] = self.info.common_network_relative_link()
                if self.info.volume_label_unicode_offset():
                    res["link_info"]["location_info"][
                        "volume_label_unicode_offset"
                    ] = self.info.volume_label_unicode_offset()
                    res["link_info"]["location_info"][
                        "volume_label_unicode"
                    ] = self.info.volume_label_unicode()
            elif type(self.info).__name__ == "Network":
                res["link_info"]["location"] = self.info.location()
                res["link_info"]["location_info"] = {
                    "common_network_relative_link_size": self.info.common_network_relative_link_size(),
                    "common_network_relative_link_flags": self.info.common_network_relative_link_flags(),
                    "net_name_offset": self.info.net_name_offset(),
                    "device_name_offset": self.info.device_name_offset(),
                    "r_network_provider_type": self.info.r_network_provider_type(),
                }
                if self.info.network_provider_type():
                    res["link_info"]["location_info"][
                        "network_provider_type"
                    ] = self.info.network_provider_type()
                if self.info.net_name_offset_unicode():
                    res["link_info"]["location_info"][
                        "net_name_offset_unicode"
                    ] = self.info.net_name_offset_unicode()
                    res["link_info"]["location_info"][
                        "net_name_unicode"
                    ] = self.info.net_name_unicode()
                if self.info.device_name_offset_unicode():
                    res["link_info"]["location_info"][
                        "device_name_offset_unicode"
                    ] = self.info.device_name_offset_unicode()
                    res["link_info"]["location_info"][
                        "device_name_unicode"
                    ] = self.info.device_name_unicode()
                if self.info.net_name():
                    res["link_info"]["location_info"]["net_name"] = self.info.net_name()
                if self.info.device_name():
                    res["link_info"]["location_info"][
                        "device_name"
                    ] = self.info.device_name()

        if not get_all:
            res["header"].pop("header_size", None)
            res["header"].pop("reserved0", None)
            res["header"].pop("reserved1", None)
            res["header"].pop("reserved2", None)
            res["link_info"].pop("link_info_size", None)
            res["link_info"].pop("link_info_header_size", None)
            res["link_info"].pop("volume_id_offset", None)
            res["link_info"].pop("local_base_path_offset", None)
            res["link_info"].pop("common_network_relative_link_offset", None)
            res["link_info"].pop("common_path_suffix_offset", None)
            if (
                "location" in res["link_info"]
                and "Local" in res["link_info"]["location"]
            ):
                res["link_info"]["location_info"].pop("volume_id_size", None)
                res["link_info"]["location_info"].pop("volume_label_offset", None)
            if (
                "location" in res["link_info"]
                and "Network" in res["link_info"]["location"]
            ):
                res["link_info"]["location_info"].pop(
                    "common_network_relative_link_size", None
                )
                res["link_info"]["location_info"].pop("net_name_offset", None)
                res["link_info"]["location_info"].pop("device_name_offset", None)

            if "target" in res:
                res["target"].pop("index", None)
                res["target"].pop("size", None)
                if "items" in res["target"]:
                    for item in res["target"]["items"]:
                        if item:
                            item.pop("modification_time", None)

        return res


def main():
    arg_parser = argparse.ArgumentParser(description=__description__)
    arg_parser.add_argument(
        dest="file",
        metavar="FILE",
        help="absolute or relative path to the file",
    )
    arg_parser.add_argument(
        "-t", "--target", action="store_true", help="print shortcut target only"
    )
    arg_parser.add_argument(
        "-j", "--json", action="store_true", help="print output in JSON"
    )
    arg_parser.add_argument(
        "-c",
        "--codepage",
        dest="cp",
        default="cp1252",
        help="set codepage of ASCII strings",
    )
    arg_parser.add_argument(
        "-a",
        "--all",
        dest="print_all",
        action="store_true",
        help="print all extracted data (i.e. offsets and sizes)",
    )
    args = arg_parser.parse_args()

    with open(args.file, "rb") as file:
        lnk = LnkFile(fhandle=file, cp=args.cp)
        if args.target:
            lnk.print_shortcut_target(pjson=args.json)
        elif args.json:
            lnk.print_json(args.print_all)
        else:
            lnk.print_lnk_file(args.print_all)


if __name__ == "__main__":
    main()
